import { QueryRunner } from "../QueryRunner";
import { ObjectLiteral } from "../../common/ObjectLiteral";
import { Logger } from "../../logger/Logger";
import { DatabaseConnection } from "../DatabaseConnection";
import { PostgresDriver } from "./PostgresDriver";
import { IndexMetadata } from "../../metadata/IndexMetadata";
import { ForeignKeyMetadata } from "../../metadata/ForeignKeyMetadata";
import { ColumnSchema } from "../../schema-builder/ColumnSchema";
import { ColumnMetadata } from "../../metadata/ColumnMetadata";
import { TableMetadata } from "../../metadata/TableMetadata";
import { TableSchema } from "../../schema-builder/TableSchema";
export declare class PostgresQueryRunner implements QueryRunner {
    protected databaseConnection: DatabaseConnection;
    protected driver: PostgresDriver;
    protected logger: Logger;
    constructor(databaseConnection: DatabaseConnection, driver: PostgresDriver, logger: Logger);
    /**
     * Releases database connection. This is needed when using connection pooling.
     * If connection is not from a pool, it should not be released.
     */
    release(): Promise<void>;
    /**
     * Removes all tables from the currently connected database.
     */
    clearDatabase(): Promise<void>;
    /**
     * Starts transaction.
     */
    beginTransaction(): Promise<void>;
    /**
     * Commits transaction.
     */
    commitTransaction(): Promise<void>;
    /**
     * Rollbacks transaction.
     */
    rollbackTransaction(): Promise<void>;
    /**
     * Checks if transaction is in progress.
     */
    isTransactionActive(): boolean;
    /**
     * Executes a given SQL query.
     */
    query(query: string, parameters?: any[]): Promise<any>;
    /**
     * Insert a new row into given table.
     */
    insert(tableName: string, keyValues: ObjectLiteral, idColumnName?: string): Promise<any>;
    /**
     * Updates rows that match given conditions in the given table.
     */
    update(tableName: string, valuesMap: ObjectLiteral, conditions: ObjectLiteral): Promise<void>;
    /**
     * Deletes from the given table by a given conditions.
     */
    delete(tableName: string, conditions: ObjectLiteral): Promise<void>;
    /**
     * Inserts rows into closure table.
     */
    insertIntoClosureTable(tableName: string, newEntityId: any, parentId: any, hasLevel: boolean): Promise<number>;
    loadSchemaTables(tableNames: string[]): Promise<TableSchema[]>;
    createTable(table: TableMetadata, columns: ColumnMetadata[]): Promise<void>;
    createColumn(tableName: string, column: ColumnMetadata): Promise<void>;
    changeColumn(tableName: string, oldColumn: ColumnSchema, newColumn: ColumnMetadata): Promise<void>;
    dropColumn(tableName: string, columnName: string): Promise<void>;
    createForeignKey(foreignKey: ForeignKeyMetadata): Promise<void>;
    dropForeignKey(tableName: string, foreignKeyName: string): Promise<void>;
    dropIndex(tableName: string, indexName: string, isGenerated?: boolean): Promise<void>;
    createIndex(tableName: string, index: IndexMetadata): Promise<void>;
    createUniqueKey(tableName: string, columnName: string, keyName: string): Promise<void>;
    normalizeType(column: ColumnMetadata): string;
    protected readonly dbName: string;
    /**
     * Parametrizes given object of values. Used to create column=value queries.
     */
    protected parametrize(objectLiteral: ObjectLiteral, startIndex?: number): string[];
    protected buildCreateColumnSql(column: ColumnMetadata, skipPrimary: boolean): string;
}
