"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const TableSchema_1 = require("./TableSchema");
const ColumnSchema_1 = require("./ColumnSchema");
const ForeignKeySchema_1 = require("./ForeignKeySchema");
const UniqueKeySchema_1 = require("./UniqueKeySchema");
const IndexSchema_1 = require("./IndexSchema");
/**
 * Creates indexes based on the given metadata.
 *
 *
 * Steps how schema creation works:
 * 1. get all tables from entity metadatas
 * 2. get all tables with complete column and keys information from the db
 * 3. drop all (old) foreign keys that exist in the table, but does not exist in the metadata
 * 4. create new tables that does not exist in the db, but exist in the metadata
 * 5. ----
 * 6. drop all columns exist (left old) in the table, but does not exist in the metadata
 * 7. add columns from metadata which does not exist in the table
 * 8. update all exist columns which metadata has changed.
 * 9. create foreign keys which does not exist in the table yet.
 * 10. create unique keys which are missing in db yet, and drops unique keys which exist in the db, but does not exist in the metadata anymore. (todo: need to move drop step to the up)
 * 11. create indices which are missing in db yet, and drops indices which exist in the db, but does not exist in the metadata anymore.
 * 12. remove primary key from the table (if it was before and does not exist in the metadata anymore).
 *
 * @internal
 */
class SchemaBuilder {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(driver, entityMetadatas) {
        this.driver = driver;
        this.entityMetadatas = entityMetadatas;
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Creates complete schemas for the given entity metadatas.
     */
    create() {
        return __awaiter(this, void 0, void 0, function* () {
            // todo: need setup connection and transaction on this level, to control parallel executions
            const metadatas = this.entityMetadatas; // todo: save to this
            this.queryRunner = yield this.driver.createQueryRunner();
            const tableSchemas = yield this.loadSchemaTables(metadatas);
            // console.log(tableSchemas);
            yield this.queryRunner.beginTransaction();
            try {
                yield this.dropOldForeignKeysForAll(metadatas, tableSchemas);
                yield this.createNewTablesForAll(metadatas, tableSchemas);
                yield this.dropRemovedColumnsForAll(metadatas, tableSchemas);
                yield this.addNewColumnsForAll(metadatas, tableSchemas);
                yield this.updateExistColumnsForAll(metadatas, tableSchemas);
                yield this.createForeignKeysForAll(metadatas, tableSchemas);
                yield this.updateUniqueKeysForAll(metadatas, tableSchemas);
                yield this.createIndicesForAll(metadatas, tableSchemas);
                yield this.removePrimaryKeyForAll(metadatas, tableSchemas);
                yield this.queryRunner.commitTransaction();
                yield this.queryRunner.release();
            }
            catch (error) {
                yield this.queryRunner.rollbackTransaction();
                yield this.queryRunner.release();
                throw error;
            }
        });
    }
    // -------------------------------------------------------------------------
    // Private Methods
    // -------------------------------------------------------------------------
    /**
     * Loads all table schemas from the database.
     */
    loadSchemaTables(metadatas) {
        const tableNames = metadatas.map(metadata => metadata.table.name);
        return this.queryRunner.loadSchemaTables(tableNames);
    }
    /**
     * Drops all (old) foreign keys that exist in the table, but does not exist in the metadata.
     */
    dropOldForeignKeysForAll(metadatas, dbTables) {
        return __awaiter(this, void 0, void 0, function* () {
            let promises = [];
            metadatas.forEach(metadata => {
                const dbTable = dbTables.find(table => table.name === metadata.table.name);
                if (!dbTable)
                    return;
                dbTable.foreignKeys
                    .filter(dbForeignKey => !metadata.foreignKeys.find(foreignKey => foreignKey.name === dbForeignKey.name))
                    .forEach(dbForeignKey => {
                    const promise = this.queryRunner
                        .dropForeignKey(metadata.table.name, dbForeignKey.name)
                        .then(() => dbTable.removeForeignKey(dbForeignKey));
                    promises.push(promise);
                });
            });
            yield Promise.all(promises);
        });
    }
    /**
     * Creates a new table if it does not exist.
     */
    createNewTablesForAll(metadatas, dbTables) {
        return __awaiter(this, void 0, void 0, function* () {
            yield Promise.all(metadatas.map((metadata) => __awaiter(this, void 0, void 0, function* () {
                const dbTable = dbTables.find(table => table.name === metadata.table.name);
                if (!dbTable) {
                    yield this.queryRunner.createTable(metadata.table, metadata.columns);
                    dbTables.push(TableSchema_1.TableSchema.createFromMetadata(this.queryRunner, metadata.table, metadata.columns));
                }
            })));
        });
    }
    /**
     * Drops all columns exist (left old) in the table, but does not exist in the metadata.
     */
    dropRemovedColumnsForAll(metadatas, dbTables) {
        const allForeignKeys = metadatas.reduce((all, metadata) => all.concat(metadata.foreignKeys), []);
        // return Promise.all(metadatas.map(metadata => this.dropRemovedColumns(metadata.table, metadata.columns, allKeys)));
        return Promise.all(metadatas.map((metadata) => __awaiter(this, void 0, void 0, function* () {
            const dbTable = dbTables.find(table => table.name === metadata.table.name);
            if (!dbTable)
                return;
            const dropColumnQueries = dbTable.columns
                .filter(dbColumn => !metadata.columns.find(column => column.name === dbColumn.name))
                .map((dbColumn) => __awaiter(this, void 0, void 0, function* () {
                yield this.dropAllColumnRelatedForeignKeys(metadata.table.name, dbColumn.name, allForeignKeys, dbTables);
                yield this.queryRunner.dropColumn(metadata.table.name, dbColumn.name);
                dbTable.removeColumn(dbColumn);
            }));
            return Promise.all(dropColumnQueries);
        })));
    }
    /**
     * Adds columns from metadata which does not exist in the table.
     */
    addNewColumnsForAll(metadatas, dbTables) {
        // return Promise.all(metadatas.map(metadata => this.addNewColumns(metadata.table, metadata.columns)));
        return Promise.all(metadatas.map((metadata) => __awaiter(this, void 0, void 0, function* () {
            const dbTable = dbTables.find(table => table.name === metadata.table.name);
            if (!dbTable)
                return;
            const newColumnQueries = metadata.columns
                .filter(column => !dbTable.columns.find(dbColumn => dbColumn.name === column.name))
                .map((column) => __awaiter(this, void 0, void 0, function* () {
                yield this.queryRunner.createColumn(metadata.table.name, column);
                dbTable.columns.push(ColumnSchema_1.ColumnSchema.create(this.queryRunner, column));
            }));
            return Promise.all(newColumnQueries);
        })));
    }
    /**
     * Update all exist columns which metadata has changed.
     */
    updateExistColumnsForAll(metadatas, dbTables) {
        const allForeignKeys = metadatas.reduce((all, metadata) => all.concat(metadata.foreignKeys), []);
        // return Promise.all(metadatas.map(metadata => this.updateExistColumns(metadata.table, metadata.columns, allKeys)));
        return Promise.all(metadatas.map((metadata) => __awaiter(this, void 0, void 0, function* () {
            const dbTable = dbTables.find(table => table.name === metadata.table.name);
            if (!dbTable)
                return;
            const updateQueries = dbTable
                .findChangedColumns(this.queryRunner, metadata.columns)
                .map((changedColumn) => __awaiter(this, void 0, void 0, function* () {
                const column = metadata.columns.find(column => column.name === changedColumn.name);
                if (!column)
                    throw new Error(`Column ${changedColumn.name} was not found in the given columns`);
                yield this.dropAllColumnRelatedForeignKeys(metadata.table.name, column.name, allForeignKeys, dbTables);
                yield this.queryRunner.changeColumn(metadata.table.name, changedColumn, column);
            }));
            return Promise.all(updateQueries);
        })));
    }
    /**
     * Creates foreign keys which does not exist in the table yet.
     */
    createForeignKeysForAll(metadatas, dbTables) {
        return Promise.all(metadatas.map((metadata) => __awaiter(this, void 0, void 0, function* () {
            const dbTable = dbTables.find(table => table.name === metadata.table.name);
            if (!dbTable)
                return;
            const dropKeysQueries = metadata.foreignKeys
                .filter(foreignKey => !dbTable.foreignKeys.find(dbForeignKey => dbForeignKey.name === foreignKey.name))
                .map((foreignKey) => __awaiter(this, void 0, void 0, function* () {
                yield this.queryRunner.createForeignKey(foreignKey);
                dbTable.foreignKeys.push(new ForeignKeySchema_1.ForeignKeySchema(foreignKey.name));
            }));
            return Promise.all(dropKeysQueries);
        })));
    }
    /**
     * Creates unique keys which are missing in db yet, and drops unique keys which exist in the db,
     * but does not exist in the metadata anymore.
     */
    updateUniqueKeysForAll(metadatas, dbTables) {
        return __awaiter(this, void 0, void 0, function* () {
            // return Promise.all(metadatas.map(metadata => this.updateUniqueKeys(metadata.table, metadata.columns)));
            yield Promise.all(metadatas.map(metadata => {
                const dbTable = dbTables.find(table => table.name === metadata.table.name);
                if (!dbTable)
                    return Promise.resolve();
                // first find metadata columns that should be unique and update them if they are not unique in db
                const addQueries = metadata.columns
                    .filter(column => column.isUnique)
                    .filter(column => !dbTable.uniqueKeys.find(uniqueKey => uniqueKey.name === "uk_" + column.name))
                    .map((column) => __awaiter(this, void 0, void 0, function* () {
                    yield this.queryRunner.createUniqueKey(metadata.table.name, column.name, "uk_" + column.name);
                    dbTable.uniqueKeys.push(new UniqueKeySchema_1.UniqueKeySchema("uk_" + column.name));
                }));
                // second find columns in db that are unique, however in metadata columns they are not unique
                const dropQueries = metadata.columns
                    .filter(column => !column.isUnique)
                    .filter(column => !!dbTable.uniqueKeys.find(uniqueKey => uniqueKey.name === "uk_" + column.name))
                    .map((column) => __awaiter(this, void 0, void 0, function* () {
                    yield this.queryRunner.dropIndex(metadata.table.name, "uk_" + column.name);
                    dbTable.removeUniqueByName("uk_" + column.name);
                }));
                return Promise.all([addQueries, dropQueries]);
            }));
        });
    }
    /**
     * Creates indices which are missing in db yet, and drops indices which exist in the db,
     * but does not exist in the metadata anymore.
     */
    createIndicesForAll(metadatas, dbTables) {
        // return Promise.all(metadatas.map(metadata => this.createIndices(metadata.table, metadata.indices)));
        return Promise.all(metadatas.map(metadata => {
            const dbTable = dbTables.find(table => table.name === metadata.table.name);
            if (!dbTable)
                return Promise.resolve();
            // drop all indices that exist in the table, but does not exist in the given composite indices
            const dropQueries = dbTable.indices
                .filter(dbIndex => !metadata.indices.find(index => index.name === dbIndex.name))
                .map((dbIndex) => __awaiter(this, void 0, void 0, function* () {
                yield this.queryRunner.dropIndex(metadata.table.name, dbIndex.name);
                dbTable.removeIndex(dbIndex);
            }));
            // then create table indices for all composite indices we have
            const addQueries = metadata.indices
                .filter(indexMetadata => !dbTable.indices.find(dbIndex => dbIndex.name === indexMetadata.name))
                .map((indexMetadata) => __awaiter(this, void 0, void 0, function* () {
                yield this.queryRunner.createIndex(metadata.table.name, indexMetadata);
                dbTable.indices.push(new IndexSchema_1.IndexSchema(indexMetadata.name, indexMetadata.columns));
            }));
            return Promise.all([dropQueries, addQueries]);
        }));
    }
    /**
     * Removes primary key from the table (if it was before and does not exist in the metadata anymore).
     */
    removePrimaryKeyForAll(metadatas, dbTables) {
        const queries = metadatas
            .filter(metadata => !metadata.hasPrimaryColumn)
            .map((metadata) => __awaiter(this, void 0, void 0, function* () {
            const dbTable = dbTables.find(table => table.name === metadata.table.name);
            if (dbTable && dbTable.primaryKey) {
                yield this.queryRunner.dropIndex(metadata.table.name, dbTable.primaryKey.name);
                dbTable.removePrimaryKey();
            }
            return undefined;
        }));
        return Promise.all(queries);
    }
    dropAllColumnRelatedForeignKeys(tableName, columnName, foreignKeys, dbTables) {
        return __awaiter(this, void 0, void 0, function* () {
            const dbTable = dbTables.find(table => table.name === tableName);
            if (!dbTable)
                return;
            // find depend foreign keys to drop them
            const dependForeignKeys = foreignKeys.filter(foreignKey => {
                if (foreignKey.tableName === tableName) {
                    return !!foreignKey.columns.find(fkColumn => {
                        return fkColumn.name === columnName;
                    });
                }
                else if (foreignKey.referencedTableName === tableName) {
                    return !!foreignKey.referencedColumns.find(fkColumn => {
                        return fkColumn.name === columnName;
                    });
                }
                return false;
            });
            if (!dependForeignKeys.length)
                return;
            yield Promise.all(dependForeignKeys.map((fk) => __awaiter(this, void 0, void 0, function* () {
                const foreignKey = dbTable.foreignKeys.find(dbForeignKey => dbForeignKey.name === fk.name);
                if (foreignKey) {
                    yield this.queryRunner.dropForeignKey(fk.tableName, fk.name);
                    dbTable.removeForeignKey(foreignKey);
                }
            })));
        });
    }
}
exports.SchemaBuilder = SchemaBuilder;

//# sourceMappingURL=SchemaBuilder.js.map
